/*
 * Copyright 2013 Sony Corporation
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
#include <common.h>
#include <command.h>

enum boot_type_id {
	normal_boot = 0,
	recovery_boot = 1,
	sdcard_boot = 3
};

#define GPIO3_OFFSET 0x2000

/*
 * SY-462: DIP-SW S4002
 * HPR-67: DIP-SW S2000
 * (on=0, off=1)
 */
#define GPIO_DIPSW_BOOTSEL_MASK     0x0c000000
#define GPIO_DIPSW_BOOTSEL_NORMAL   (~0x04000000 & GPIO_DIPSW_BOOTSEL_MASK)
#define GPIO_DIPSW_BOOTSEL_RECOVERY (~0x08000000 & GPIO_DIPSW_BOOTSEL_MASK)
static inline uint32_t read_dipsw_bootsel(void)
{
	ccsr_gpio_t __iomem *pgpio = (void *)(CONFIG_SYS_MPC85xx_GPIO_ADDR + GPIO3_OFFSET);
	u32 val = in_be32(&pgpio->gpdat);
	debug("%08x %08x at %p\n", pgpio->gpdat, in_be32(&pgpio->gpdat), pgpio);
	return val & GPIO_DIPSW_BOOTSEL_MASK;
}

#define NVSYS_BOOTMODE_OFFSET 0 /* 2byte */
#define BOOTMODE_NVRAM_NORMAL_MAGIC 0x2549
static inline enum boot_type_id get_auto_boot_type(void)
{
	enum boot_type_id type = recovery_boot;

#ifdef CONFIG_SYS_NVSRAM_BASE
	volatile uint16_t *p_magic = (volatile uint16_t *)(CONFIG_SYS_NVSRAM_BASE + NVSYS_BOOTMODE_OFFSET);

	uint16_t value = in_be16(p_magic);

	debug("%s: value=%04x\n", __func__, value);

	if (value == BOOTMODE_NVRAM_NORMAL_MAGIC) {
		type = normal_boot;
	}
#endif

	return type;
}

int do_bootselect(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	const char *normal_str = "normal";
	const char *recovery_str = "recovery";
	const char *sdcard_str = "sdcard";
	char *bootargs_p;
	char *bootcmd_p;

	int boot_type;

	if (argc > 2) {
		puts("Too many arguments.\n");
		return 1;
	}

	if (argc == 2) {		/* User select */
		if (strncmp(normal_str, argv[1], strnlen(normal_str,32)) == 0) {
			boot_type = normal_boot;
		} else if  (strncmp(recovery_str, argv[1], strnlen(recovery_str,32)) == 0) {
			boot_type = recovery_boot;
		} else if  (strncmp(sdcard_str, argv[1], strnlen(sdcard_str,32)) == 0) {
			boot_type = sdcard_boot;
		} else {
			puts("Illegal argument.\n");
			return 1;
		}
	} else if (argc == 1) {		/* Auto select */
#ifdef CONFIG_SDCARD
		boot_type = sdcard_boot;
#else /* CONFIG_SDCARD */
#ifdef CONFIG_SONY_BOARD
		uint32_t bootsel_val = read_dipsw_bootsel();

		switch (bootsel_val) {
		case GPIO_DIPSW_BOOTSEL_NORMAL:
			/* force normal */
			boot_type = normal_boot;
			break;
		case GPIO_DIPSW_BOOTSEL_RECOVERY:
			/* force recovery */
			boot_type = recovery_boot;
			break;
		default:
			/* auto select */
			boot_type = get_auto_boot_type();
			break;
		}
#else /* CONFIG_SONY_BOARD */
		boot_type = normal_boot;
#endif /* CONFIG_SONY_BOARD */
#endif /* CONFIG_SDCARD */
	} else {
		puts("Internal error.\n");
		return 1;
	}

	switch (boot_type) {
	case normal_boot:
		puts("Normal boot mode\n");
#ifdef CONFIG_SPIFLASH
		bootargs_p = getenv("bootargs_spi_n");
		bootcmd_p = getenv("bootcmd_spi_n");
#else
		bootargs_p = getenv("bootargs_n");
		bootcmd_p = getenv("bootcmd_n");
#endif
		break;
	case recovery_boot:
		puts("Recovery boot mode\n");
#ifdef CONFIG_SPIFLASH
		bootargs_p = getenv("bootargs_spi_r");
		bootcmd_p = getenv("bootcmd_spi_r");
#else
		bootargs_p = getenv("bootargs_r");
		bootcmd_p = getenv("bootcmd_r");
#endif
		break;
	case sdcard_boot:
		puts("SDCARD boot mode\n");
		bootargs_p = getenv("bootargs_sd");
		bootcmd_p = getenv("bootcmd_sd");
		break;
	default:
		puts("Internal Error\n");
		return 1;
	}

	setenv("bootargs", bootargs_p);
	setenv("bootcmd", bootcmd_p);

	return 0;
}

U_BOOT_CMD(
	bootselect,	2,	1,	do_bootselect,
	"change bootmode",
	"[normal|recovery|sdcard]"
);
